//@ts-check
/**
 * This is the version that uses ts-morph but it is rather slow right now,
 * So we are switching to fixApiModelImportsV2.js which uses native typescript compiler api.
 */

/**
 * @typedef {(p: string)=>string} RelativePathSpecifierFunc
 */

const fs = require("fs-extra");
const _ = require("lodash");
const path = require("path");
const assert = require("assert");
const util = require("util");
const chalk = require("chalk");
const open = require("open");
const globby = require("globby");
const {
	Project,
	Directory,
	Diagnostic,
	SourceFile,
	ts,
	printNode,
	StructureKind,
} = require("ts-morph");

const debug0 = (/** @type {any} */ obj) => util.inspect(obj, { depth: 0 });
const debug1 = (/** @type {any} */ obj) => util.inspect(obj, { depth: 1 });

/** @type {(n: ts.SyntaxKind)=> string} */
const formatSyntaxKind = (kind) => ts.SyntaxKind[kind];

/** @type {(n: StructureKind)=> string} */
const formatStructureKind = (kind) => StructureKind[kind];

/**
 * @param {number} ms 
 */
function sleep(ms) {
	return new Promise((res, rej) => setTimeout(res, ms));
}

/**
 * Remove repeated enum declarations, move them to separate file and import them
 * @param {Project} project
 * @param {RelativePathSpecifierFunc} relativeModuleSpecifier
 * @param {SourceFile} src
 */
function fixInlineEnumDeclarations(project, src, relativeModuleSpecifier) {
	const enumsDecls = src.getEnums();
	const srcPath = src.getFilePath();
	for (const enumDecl of enumsDecls) {
		const standaloneName = enumDecl.getName() + ".ts";
		const standalonePath = path.join(path.dirname(srcPath), standaloneName);
		const found = project.getSourceFile(standalonePath);
		if (!found) {
			console.log("Adding", standaloneName);
			project.createSourceFile(standalonePath, {
				kind: StructureKind.SourceFile,
				statements: [enumDecl.getStructure()],
			});
		}
	}
	const decls = _.uniq(enumsDecls.map(e => e.getName()));
	if (decls.length) {
		src.addImportDeclarations(decls.map(n => ({
			namedImports: [n],
			moduleSpecifier: relativeModuleSpecifier(`model/${n}`),
		})));
		enumsDecls.forEach(e => e.remove());
	}
}

/**
 * fix type references like: `models.CoreRegionInputModel` by removing models and add imports
 * @param {Project} project
 * @param {RelativePathSpecifierFunc} relativeModuleSpecifier
 * @param {SourceFile} src
 */
function fixReferencingModelsDot(project, src, relativeModuleSpecifier) {
	const refs = src.getDescendantsOfKind(ts.SyntaxKind.TypeReference)
	/** @type {string[]} */
	const referencedModelTypes = [];
	for (const ref of refs) {
		const typeName = ref.getTypeName();
		if (ts.isQualifiedName(typeName.compilerNode)) {
			typeName.transform(traversal => {
				const node = traversal.visitChildren();
				if (ts.isQualifiedName(node)) {
					const left = node.left.getText();
					const modelType = node.right.getText();
					if (left === "models") {
						referencedModelTypes.push(modelType);
						return ts.factory.createTypeReferenceNode(modelType);
					}
				}
				return node;
			});
		}
	}
	const currentModuleName = path.basename(src.getFilePath()).replace(/\.ts$/, "");
	const imports = _.uniq(referencedModelTypes).filter(r => r !== currentModuleName);
	// console.log("removing", currentModuleName, "out of", imports);
	if (imports.length) {
		src.addImportDeclarations(imports.map(n => ({
			namedImports: [n],
			moduleSpecifier: relativeModuleSpecifier(`model/${n}`),
		})));
	}
}

/**
 * fix type references like: `models.CoreRegionInputModel` by removing models and add imports
 * @param {Project} project
 * @param {RelativePathSpecifierFunc} relativeModuleSpecifier
 * @param {SourceFile} src
 * @param {Diagnostic[]} diags
 */
function fixMissingRefs(project, src, relativeModuleSpecifier, diags) {
	/** @type {string[]} */
	const missingRefs = [];
	for (const diag of diags) {
		if (diag.getCode() === CODE_CAN_NOT_FIND_NAME) {
			const start = diag.getStart();
			assert(start, "start falsy");
			const child = src.getDescendantAtPos(start);
			assert(child, "child falsy");
			if (ts.isIdentifier(child.compilerNode)) {
				console.log(child.getKindName(), child.print(), diag.getMessageText());
				var ident = child.compilerNode.text;
				missingRefs.push(ident);
			}
		}
	}
	const imports = _.uniq(missingRefs);
	if (imports.length) {
		src.addImportDeclarations(imports.map(n => ({
			namedImports: [n],
			moduleSpecifier: relativeModuleSpecifier(`model/${n}.ts`),
		})));
	}
}

const CODE_CAN_NOT_FIND_NAME = 2304;

/**
 * @param {string} apisDirectory
 */
async function fixApiModelImports(apisDirectory) {
	console.time("total");
	const toBeRemoved = ["model/models.ts", "variables.ts", "index.ts", "configuration.ts"];
	for (const r of toBeRemoved) {
		const p = path.join(apisDirectory, r);
		if (await fs.pathExists(p)) {
			await fs.remove(p);
		}
	}
	assert(fs.pathExists(apisDirectory), `${apisDirectory} does not exist in disk.`);
	const modelFiles = await globby("**/*.ts", { cwd: apisDirectory, absolute: true });

	assert(modelFiles.length, "gotta have a few models");
	const project = new Project();
	project.addSourceFilesAtPaths(modelFiles);
	const rootDirectory = project.getDirectory(apisDirectory);
	assert(rootDirectory, "rootDirectory falsy");
	const rootDirectoryPath = rootDirectory.getPath();

	/**
	 * @param {Directory} srcDirectory 
	 */
	function makeRelativeModuleSpecifier(srcDirectory) {
		assert(srcDirectory, "srcDirectory falsy");
		/** @type {RelativePathSpecifierFunc} */
		const func = (relativePath) => {
			assert(relativePath, "relativePath falsy");
			const fullPath = path.join(rootDirectoryPath, relativePath);
			return srcDirectory.getRelativePathAsModuleSpecifierTo(fullPath);
		}
		return func;
	}

	// project.resolveSourceFileDependencies();
	const srcs = project.getSourceFiles();
	const totalCount = srcs.length;
	console.time("step1");
	let counter = 0;
	for (const src of srcs) {
		counter++;
		const p = src.getFilePath();
		const filename = path.basename(p);
		const isEnumFile = filename.endsWith("Enum.ts");
		if (isEnumFile) continue;

		const dirname = path.basename(path.dirname(p));
		if (dirname === "model") {
			console.time("fix1");
			fixInlineEnumDeclarations(project, src, makeRelativeModuleSpecifier(src.getDirectory()));
			console.timeEnd("fix1");
		}
		console.time("fix2");
		fixReferencingModelsDot(project, src, makeRelativeModuleSpecifier(src.getDirectory()));
		console.timeEnd("fix2");
		await sleep(1);
	}
	console.timeEnd("step1");

	console.time("step3");
	const allDiags = _.groupBy(project.getPreEmitDiagnostics(), d => {
		const src = d.getSourceFile();
		assert(src, "gotta have src");
		return src.getFilePath();
	});
	console.timeEnd("step3");

	console.time("step4");
	counter = 0;
	for (const src of srcs) {
		counter++;
		const diags = allDiags[src.getFilePath()] || [];
		fixMissingRefs(project, src, makeRelativeModuleSpecifier(src.getDirectory()), diags);
		await sleep(1);
	}
	console.timeEnd("step4");

	console.time("step5");
	await project.save();
	const diagnostics = project.getPreEmitDiagnostics();
	console.log(project.formatDiagnosticsWithColorAndContext(diagnostics));
	console.timeEnd("step5");
	console.timeEnd("total");
}

module.exports = {
	fixApiModelImports,
};

if (require.main === module) {
	(async () => {
		await fixApiModelImports("./example_code_gen_for_jquery_v2");
	})();
}