//@ts-check
/**
 * This is the version that uses ts-morph but it is rather slow right now,
 * So we are switching to fixApiModelImportsV2.js which uses native typescript compiler api.
 */

/**
 * @typedef {import("typescript").Node} Node
 * @typedef {import("typescript").SourceFile} SourceFile
 * @typedef {import("typescript").SyntaxKind} SyntaxKind
 * @typedef {import("typescript").Statement} Statement
 * @typedef {import("typescript").ImportDeclaration} ImportDeclaration
 * @typedef {import("typescript").FunctionDeclaration} FunctionDeclaration
 * @typedef {import("typescript").TransformerFactory<SourceFile>} TransformerFactory
 */

const fs = require("fs-extra");
const _ = require("lodash");
const ts = require("typescript");
const path = require("path");
const assert = require("assert");
const util = require("util");
const chalk = require("chalk");
const globby = require("globby");
const { Project } = require("ts-morph");
const { ScriptKind } = require("typescript");

const debug0 = (/** @type {any} */ obj) => util.inspect(obj, { depth: 0 });
const debug1 = (/** @type {any} */ obj) => util.inspect(obj, { depth: 1 });

const nodePrinter = ts.createPrinter({ newLine: ts.NewLineKind.LineFeed });

/** @type {(src: SourceFile)=>string} */
const printSourceFile = (src) => nodePrinter.printNode(ts.EmitHint.SourceFile, src, src);

const DEBUG_FILE_NAME = process.env.DEBUG_FILE_NAME;
const NO_TRANSFORM = Boolean(process.env.NO_TRANSFORM);

/** @type {(kind: SyntaxKind)=>string} */
const formatSyntaxKind = (kind) => ts.SyntaxKind[kind];

/**
 * @param {string} packagePath
 * @param {string[]} vars
 * @returns {ImportDeclaration}
 */
function makeImportStatement(packagePath, vars) {
	return ts.createImportDeclaration(
		/* decorators */ undefined,
		/* modifiers */ undefined,
		ts.createImportClause(undefined,
			ts.createNamedImports(
				vars.map(v => ts.createImportSpecifier(undefined, ts.createIdentifier(v))))
		),
		ts.createLiteral(packagePath)
	);
}


/**
 * @param {SourceFile} sc
 * @returns {SourceFile}
 */
function fixValidationRegex(sc) {
	assert(sc, "sc falsy");

	/** @type {TransformerFactory} */
	const removeModelsPrefixTransformer = (context) => {
		return (sourceFile) => {
			/**
			 * @param {Node} node
			 * @returns {Node | Node[] | undefined}
			 */
			function visitor(node) {
				if (ts.isRegularExpressionLiteral(node)) {
					const text = node.text;
					return ts.createRegularExpressionLiteral(text.replace(/\\\\/g, '\\'));
				}
				return ts.visitEachChild(node, visitor, context);
			}
			return ts.visitNode(sourceFile, visitor);
		}
	};

	const res = ts.transform(sc, [removeModelsPrefixTransformer]);
	const transformed = res.transformed[0];
	return transformed;
}

/**
 * @param {SourceFile} sc
 * @param {Set<string>} allTsFiles
 * @returns {SourceFile}
 */
function removeInlineEnumsAndAddingImports(sc, allTsFiles) {
	assert(sc, "sc falsy");
	/** @type {string[]} */
	const inlineEnumNames = [];
	//remove inline enums
	ts.forEachChild(sc, (node) => {
		if (ts.isEnumDeclaration(node)) {
			const enumName = node.name.text;
			assert(enumName && enumName.endsWith("Enum"), `invalid enum name: ${enumName} in: ${node.getText(sc)}`);
			inlineEnumNames.push(enumName);
			const expectedEnumPath = `model/${enumName}.ts`;
			assert(allTsFiles.has(expectedEnumPath), `${expectedEnumPath} not defined`);
		}
	});
	const transformed = ts.updateSourceFileNode(sc, [
		...inlineEnumNames.map(e => makeImportStatement(`../model/${e}`, [e])),
		...sc.statements.filter(s => !ts.isEnumDeclaration(s))
	]);
	return transformed;
}

/**
 * @param {SourceFile} sc
 * @returns {SourceFile}
 */
function removeModelsPrefixAndAddImports(sc) {
	assert(sc, "sc falsy");
	/** @type {string[]} */
	const referencedModelNames = [];

	/** @type {TransformerFactory} */
	const removeModelsPrefixTransformer = (context) => {
		return (sourceFile) => {
			/**
			 * @param {Node} node
			 * @returns {Node | Node[] | undefined}
			 */
			function visitor(node) {
				if (ts.isQualifiedName(node)) {
					if (ts.isIdentifier(node.left) && node.left.getText() === "models") {
						// console.log("removing models. prefix:", node.getText());
						const modelName = node.right.getText();
						referencedModelNames.push(modelName)
						return ts.createIdentifier(modelName);
					}
				}
				return ts.visitEachChild(node, visitor, context);
			}
			return ts.visitNode(sourceFile, visitor);
		}
	};

	const res = ts.transform(sc, [removeModelsPrefixTransformer]);
	const transformed = res.transformed[0];
	const transformed2 = ts.updateSourceFileNode(transformed, [
		...referencedModelNames.map(e => makeImportStatement(`../model/${e}`, [e])),
		...transformed.statements
	]);

	// const code = printSourceFile(transformed2);
	// if (code.includes("models.")) {
	// 	throw new Error(`still have models. in file: ${sc.fileName}`);
	// }
	return transformed2;
}

/**
 * @param {ImportDeclaration} imp
 * @return {string}
 */
function getImportPath(imp) {
	assert(imp, "imp falsy");
	const spec = imp.moduleSpecifier;
	assert(spec, "spec falsy");
	assert(ts.isStringLiteral(spec), "spec gotta be string literal but is: " + formatSyntaxKind(spec.kind));
	const p = spec.text;
	assert(p, "p falsy");
	return p;
}

/**
 * @param {ImportDeclaration} imp
 * @return {Set<string>}
 */
function getNamedImports(imp) {
	assert(imp, "imp falsy");
	/** @type {Set<string>} */
	const namedImports = new Set();
	const clause = imp.importClause;
	if (clause && clause.namedBindings) {
		const bindings = clause.namedBindings;
		if (ts.isNamedImports(bindings)) {
			bindings.elements.map(el => el.name.text).forEach(name => namedImports.add(name));
		}
	}
	return namedImports;
}

/**
 * @param {SourceFile} sc
 * @param {Set<string>} justMyTypes
 * @returns {SourceFile}
 */
function fixMissingAndDuplicateAndUnusedReferences(sc, justMyTypes) {
	assert(sc, "sc falsy");
	assert(justMyTypes, "justMyTypes falsy");
	assert(justMyTypes.size, "justMyTypes empty");
	const DEBUG_FILE = sc.fileName.endsWith(DEBUG_FILE_NAME);

	/** @type {Set<string>} */
	const allReferencedTypes = new Set();
	/** @type {Set<string>} */
	const allIdentifiers = new Set();
	/**
	 * @param {Node} node 
	 */
	function typeReferenceVisitor(node) {
		if (ts.isTypeReferenceNode(node)) {
			const typeName = node.typeName;
			if (ts.isIdentifier(typeName)) {
				if (!node.typeArguments) {
					if (!typeName.text) {
						let { line } = sc.getLineAndCharacterOfPosition(node.getStart(sc));
						throw new Error(`typeName.text is falsy in node: ${node.getText(sc)} in ${sc.fileName}:${line}`);
					}
					const refId = typeName.text;
					assert(refId, "refId falsy")
					if (justMyTypes.has(refId)) {
						allReferencedTypes.add(refId);
					}
				}
			}
		} else if (ts.isExpressionWithTypeArguments(node)) {
			const expr = node.expression;
			if (ts.isIdentifier(expr)) {
				assert(expr.text, "expr.text falsy");
				allReferencedTypes.add(expr.text);
			}
		} else if (ts.isDecorator(node)) {
			const ce = node.expression;
			if (ts.isCallExpression(ce)) {
				const id = ce.expression;
				if (ts.isIdentifier(id)) {
					assert(id.text, "id.text falsy");
					allReferencedTypes.add(id.text);
				}
			}
		} else if (ts.isIdentifier(node)) {
			if (DEBUG_FILE) {
				if (node.text === "Min") {
					console.log("NODE:", chalk.bgRed(printSourceFile(sc)));
					console.log(chalk.bgGreen(printSourceFile(node.getSourceFile())));
				}
			}
			allIdentifiers.add(node.text);
		}
		node.forEachChild(typeReferenceVisitor);
	}
	sc.forEachChild(typeReferenceVisitor);
	/** @type {ImportDeclaration[]} */
	const allImports = [];
	/** @type {Set<string>} */
	const allImportedVars = new Set();
	/** @type {Set<string>} */
	const typesDefinedLocally = new Set();
	for (const node of sc.statements) {
		if (ts.isImportDeclaration(node)) {
			allImports.push(node);
			for (const namedImport of getNamedImports(node)) {
				allImportedVars.add(namedImport);
			}
		} else if (ts.isClassDeclaration(node) || ts.isInterfaceDeclaration(node)) {
			const name = node.name;
			assert(name, "name falsy");
			typesDefinedLocally.add(name.text);
		}
	}

	const notImported = _.difference(Array.from(allReferencedTypes), Array.from(allImportedVars));
	const notUsedImports = _.difference(Array.from(allImportedVars), Array.from(allIdentifiers));

	if (DEBUG_FILE) {
		console.log(chalk.bgRed.black(`============= ${sc.fileName} =================`));
		console.log(chalk.yellow(printSourceFile(sc)));
		console.log(`----------------------------------`);
		console.log("allIdentifiers", allIdentifiers);
		console.log("allReferencedTypes", allReferencedTypes);
		console.log("allImportedVars", allImportedVars);
		console.log("----------------------------------");
		console.log("notImported", notImported);
		console.log("notUsedImports", notUsedImports);
		console.log(`----------------------------------\r\n`);
	}

	const fixedImports = _.chain(allImports)
		.concat(notImported.map(e => makeImportStatement(`../model/${e}`, [e])))
		.uniqBy(getImportPath)
		.flatMap(imp => {
			const namedImports = Array.from(getNamedImports(imp));
			if (_.intersection(namedImports, Array.from(typesDefinedLocally)).length) {
				//in previous steps, we might added an import to an type that is defined locally;
				return [];
			}
			const unusedImports = _.intersection(namedImports, notUsedImports);
			if (!unusedImports.length) {
				return [imp];
			}

			if (unusedImports.length === namedImports.length) {
				//none of the imported variables are used, so this import statement can be completedly removed
				return [];
			}

			if (!imp.importClause || !imp.importClause.namedBindings) {
				return [imp];
			}

			const namedBindings = imp.importClause.namedBindings;
			if (!ts.isNamedImports(namedBindings)) {
				return [imp];
			}

			const updatedNamedImports = ts.updateNamedImports(namedBindings,
				namedBindings.elements.filter(el => !unusedImports.includes(el.name.text)));
			const impWithUnusedRemoved = ts.updateImportDeclaration(imp, undefined, undefined,
				ts.updateImportClause(imp.importClause, undefined,
					updatedNamedImports,
					imp.importClause.isTypeOnly)
				, imp.moduleSpecifier);
			return [impWithUnusedRemoved];
		}).value();

	const transformed = ts.updateSourceFileNode(sc, [
		...fixedImports,
		...sc.statements.filter(s => !ts.isImportDeclaration(s))
	]);
	return transformed;
}
/**
 * @param {SourceFile} sc
 * @returns {Promise<SourceFile>}
 */
async function moveEnumUtilsToSeparateFiles(sc) {
	assert(sc, "sc falsy");
	/** @type {FunctionDeclaration[]} */
	const funcDecls = [];

	/** @type {Statement[]} */
	const otherStmts = [];

	for (const s of sc.statements) {
		if (ts.isFunctionDeclaration(s)) {
			funcDecls.push(s);
		} else {
			otherStmts.push(s);
		}
	}
	for (const func of funcDecls) {
		assert(func.name, `func.name falsy in ${sc.fileName}`);
		const funcName = func.name.text;
		const enumName = path.basename(sc.fileName).replace(/\.ts$/, "");
		assert(funcName, `funcName falsy in ${sc.fileName}`)
		const fileName = path.join(path.dirname(sc.fileName), funcName + ".ts");
		const code = `
import { ${enumName} } from "./${enumName}"
${func.getText(sc)}
		`;
		await fs.writeFile(fileName, code, { encoding: "utf-8" });
	}

	const transformed = ts.updateSourceFileNode(sc, otherStmts);
	return transformed;
}

/**
 * @param {string} apisDirectory
 */
async function fixApiModelImportsV2(apisDirectory) {
	console.time("total");
	/** @type {(r:string)=> string} */
	const resolve = r => path.join(apisDirectory, r);
	if (await fs.pathExists(resolve("variables.ts"))) {
		await fs.rename(resolve("variables.ts"), resolve("globals.d.ts"));
	}
	if (await fs.pathExists(resolve("configuration.ts"))) {
		await fs.rename(resolve("configuration.ts"), resolve("tsconfig.json"));
	}
	// if (await fs.pathExists(resolve("index.ts"))) {
	// 	await fs.rename(resolve("index.ts"), resolve("api/_customValidators.ts"));
	// }
	const toBeRemoved = ["model/models.ts", "index.ts"];
	for (const r of toBeRemoved) {
		const p = resolve(r);
		if (await fs.pathExists(p)) {
			await fs.remove(p);
		}
	}

	assert(fs.pathExists(apisDirectory), `${apisDirectory} does not exist in disk.`);
	const _allTsFiles = await globby("**/*.ts", { cwd: apisDirectory, absolute: false });
	const apiTsFile = _allTsFiles.find(f => path.basename(f) === "api.ts");
	assert(apiTsFile, "gotta have apiTsFile");
	const apiTsFullPath = path.join(apisDirectory, apiTsFile);
	const p = new Project();
	p.addSourceFileAtPath(apiTsFullPath);
	const apiSc = p.getSourceFile(apiTsFullPath);
	assert(apiSc, "apiSc falsy");

	if (NO_TRANSFORM) {
		console.log(chalk.bgRed("NO TRANSFORM. EXIT AND OPEN DIRECTORY"));
		const open2 = require("open");
		await open2(apisDirectory, { app: { name: "code" }, wait: true });
		process.exit(1);
	}

	const apiTsExportedVars = Array.from(apiSc.getExportedDeclarations().keys());
	assert(apiTsExportedVars.length, "api.ts gotta exported some vars");

	const allTsFiles = new Set(_allTsFiles);
	const justMyTypes = new Set(_allTsFiles.filter(f => path.dirname(f).endsWith("model"))
		.map(f => path.basename(f).replace(/\.ts$/, ""))
		.concat(apiTsExportedVars)
		.concat(["customValidators"])
	);

	/** @type {(s:string)=>string} */
	const header = (s) => chalk.bgMagenta(`===================== ${s} ===================\r\n`);
	for (const tsFile of allTsFiles) {
		const DEBUG_FILE = tsFile.endsWith(DEBUG_FILE_NAME);
		const tsFullPath = path.join(apisDirectory, tsFile);
		const code = await fs.readFile(tsFullPath, { encoding: "utf-8" });
		const fileName = path.basename(tsFile);
		if (fileName === "api.ts") continue; //do not touch api.ts as well;
		const sc = ts.createSourceFile(tsFullPath, code, ts.ScriptTarget.Latest, true, ScriptKind.TS);
		const isEnum = fileName.endsWith("Enum.ts");
		let transformed = sc;
		if (DEBUG_FILE) console.log(`${header("ORIGINAL")}${tsFullPath}\r\n${chalk.yellow(code)}`);
		if (DEBUG_FILE) console.log(`${header("FIRST")}${printSourceFile(transformed)}`);
		if (isEnum) {
			transformed = await moveEnumUtilsToSeparateFiles(transformed);
			if (DEBUG_FILE) console.log(`${header("STEP 1 AFTER moveEnumUtilsToSeparateFiles")}${printSourceFile(transformed)}`);
		} else {
			transformed = fixValidationRegex(transformed);
			transformed = removeInlineEnumsAndAddingImports(transformed, allTsFiles);
			if (DEBUG_FILE) console.log(`${header("STEP 2 AFTER removeInlineEnumsAndAddingImports")}${printSourceFile(transformed)}`);
			transformed = removeModelsPrefixAndAddImports(transformed);
			if (DEBUG_FILE) console.log(`${header("STEP 3 AFTER removeModelsPrefixAndAddImports")}${printSourceFile(transformed)}`);
			transformed = fixMissingAndDuplicateAndUnusedReferences(transformed, justMyTypes);
			if (DEBUG_FILE) console.log(`${header("STEP 4 AFTER fixMissingAndDuplicateAndUnusedReferences")}${printSourceFile(transformed)}`);
		}

		const output = printSourceFile(transformed);
		// console.log("=======================================");
		// console.log(output);
		// console.log("Writing", tsFullPath);
		if (DEBUG_FILE) {
			console.log(chalk.bgRed(`DEBUGGING FILE: ${DEBUG_FILE_NAME}, EXIT EARLY AND OPEN DIRECTORY`));
			const open = require("open");
			await open(apisDirectory, { app: { name: "code" } });
			process.exit(1);
		}
		await fs.writeFile(tsFullPath, output, { encoding: "utf-8" });
	}
	// return;
	console.log("type checking...");
	const tsConfigFilePath = resolve("tsconfig.json");
	assert(await fs.pathExists(tsConfigFilePath), `${tsConfigFilePath} does not exist`);
	const project = new Project({
		tsConfigFilePath,
	});
	project.addSourceFilesFromTsConfig(tsConfigFilePath);
	// project.addSourceFilesAtPaths(Array.from(allTsFiles).map(f => path.join(apisDirectory, f)));
	const diags = project.getPreEmitDiagnostics();
	console.log(project.formatDiagnosticsWithColorAndContext(diags));
	if (!diags.length) {
		console.log(chalk.bgGreen("Success! No type errors found."));
	} else {
		console.log(chalk.bgRed(`Error: ${diags.length} type errors found.`));
	}
}

module.exports = {
	fixApiModelImportsV2,
};

if (require.main === module) {
	(async () => {
		await fixApiModelImportsV2("./example_code_gen_for_jquery_v2");
	})();
}