//@ts-check

const fs = require("fs");
const _ = require("lodash");
const path = require("path");
const assert = require("assert");
const ts = require("typescript");
const util = require("util");
const chalk = require("chalk");
/**
 * @typedef {import("typescript").TransformerFactory<import("typescript").SourceFile>} TransformerFactory
 * @typedef {import("typescript").Node} Node
 */

/**
 * @param {string} filePath
 */
async function fixUseBeforeDeclaration(filePath) {
	const fileName = path.basename(filePath);
	const code = await fs.promises.readFile(filePath, { encoding: "utf-8" });
	//see https://github.com/Microsoft/TypeScript/wiki/Using-the-Compiler-API
	//also https://github.com/Microsoft/TypeScript/issues?utf8=%E2%9C%93&q=is%3Amerged%20label%3A%22Domain%3A%20Transforms%22%20
	//also https://github.com/microsoft/TypeScript/pull/13940
	//also https://ts-ast-viewer.com/
	//also https://blog.scottlogic.com/2017/05/02/typescript-compiler-api-revisited.html
	//also https://github.com/madou/typescript-transformer-handbook
	const sc = ts.createSourceFile(fileName, code, ts.ScriptTarget.Latest, true);
	const printer = ts.createPrinter({ newLine: ts.NewLineKind.LineFeed });
	/** @type {{[k:string]:{dependencies:string[], pos: number}}} */
	const dependenciesMap = {};
	/** @type {{[k:string]:Node}} */
	const classNodesMap = {};
	ts.forEachChild(sc, (node) => {
		if (ts.isClassDeclaration(node)) {
			//see https://stackoverflow.com/questions/42801954/how-can-use-type-checker-get-interface-or-class-extends-type
			/** @type {string[]} */
			let dependencies = [];
			if (node.heritageClauses) {
				for (const hc of node.heritageClauses) {
					if (hc.token === ts.SyntaxKind.ExtendsKeyword) {
						for (const t of hc.types) {
							const baseName = printer.printNode(ts.EmitHint.Unspecified, t, sc);
							dependencies.push(baseName);
						}
					}
				}
			}
			assert(node.name, "gotta have node.name");
			const className = node.name.escapedText.toString();

			/** @type {(node: Node)=>void} */
			const visitor = (node) => {
				if (ts.isPropertyAccessExpression(node)) {
					const name = node.name;
					if (ts.isIdentifier(name) && name.text === "VALIDATORS") {
						const expr = node.expression;
						if (ts.isIdentifier(expr)) {
							dependencies.push(expr.text);
						}
						// if (ts.isPropertyAssignment) console.log("CODE: ", node.initializer.kind, s);
					}
				}
				ts.forEachChild(node, visitor);
			};

			ts.forEachChild(node, visitor);

			dependenciesMap[className] = {
				dependencies,
				pos: node.pos,
			};
			classNodesMap[className] = node;
		}
	});

	/** @type {{[k:string]:string[]}} */
	const useBeforeDeclarations = {};
	for (const [k, v] of Object.entries(dependenciesMap)) {
		if (v.dependencies.length) {
			// console.log(k, ":\r\n\t\t", v);
			for (const dep of v.dependencies) {
				const depFound = dependenciesMap[dep];
				if (depFound && depFound.pos > v.pos) {
					if (!useBeforeDeclarations[k]) {
						useBeforeDeclarations[k] = [];
					}
					useBeforeDeclarations[k].push(dep);
				}
			}
		}
	}

	// console.log("Use before declarations:", useBeforeDeclarations);
	const allUsedBasesBeforeDeclaration = _.uniq(_.flatMap(Object.values(useBeforeDeclarations)));
	/** @type {string[]} */
	const alreadyOutputedClasses = [];

	/**
	 * @param {string[]} deps
	 * @returns {IterableIterator<string>}
	 */
	function* findDepsRecursively(deps) {
		for (const dep of deps) {
			const d = dependenciesMap[dep];
			if (d.dependencies && d.dependencies.length) {
				const ds = findDepsRecursively(d.dependencies);
				yield* ds;
			}
			yield dep;
		}
	}

	/** @type {TransformerFactory} */
	const fixClassOrderTransformer = (context) => {
		return (sourceFile) => {
			/**
			 * @param {Node} node
			 * @returns {Node | Node[]}
			 */
			function visitor(node) {
				if (ts.isClassDeclaration(node)) {
					assert(node.name, "node.name falsy");
					const className = node.name.escapedText.toString();
					const usedDepsBeforeDeclaration = useBeforeDeclarations[className] || [];
					const recursiveDeps = Array.from(findDepsRecursively(usedDepsBeforeDeclaration));
					const classesPrepared = recursiveDeps.concat([className]);
					const toBeOutputed = _.uniq(_.difference(classesPrepared, alreadyOutputedClasses));
					toBeOutputed.forEach((c) => alreadyOutputedClasses.push(c));
					return toBeOutputed.map((d) => classNodesMap[d]);
				}
				return ts.visitEachChild(node, visitor, context);
			}
			return ts.visitNode(sourceFile, visitor);
		};
	};

	/** @type {TransformerFactory} */
	const fixRegexDoubleSlashTransformer = (context) => {
		return (sourceFile) => {
			/**
			 * @param {Node} node
			 * @returns {ts.Node}
			 */
			function visitor(node) {
				if (ts.isRegularExpressionLiteral(node)) {
					return ts.createRegularExpressionLiteral(node.text.replace(/\\\\/g, "\\"));
				}
				return ts.visitEachChild(node, visitor, context);
			}
			return ts.visitNode(sourceFile, visitor);
		};
	};

	const res = ts.transform(sc, [fixClassOrderTransformer, fixRegexDoubleSlashTransformer]);

	const output = printer.printNode(ts.EmitHint.SourceFile, res.transformed[0], sc);

	const diagnostics = res.diagnostics || [];
	// console.log("===========================");
	if (diagnostics.length) {
		console.log(chalk.bgRed(`${diagnostics.length} errors found`));
		for (const d of diagnostics) {
			console.log(chalk.red(`TS${d.code}`), d.messageText, `check file: ${filePath}`);
		}
		process.exit(1);
	} else {
		// const backupFilePath = filePath.replace(".ts", ".bak.ts");
		// await fs.promises.copyFile(filePath, backupFilePath);
		// console.log("backup file at: ", backupFilePath);
		await fs.promises.writeFile(filePath, output, { encoding: "utf-8" });
		console.log(chalk.bgGreen("DONE without errors"));
	}
}

module.exports = {
	fixUseBeforeDeclaration,
};
