#!/usr/bin/env node
//@ts-check

/**
 * @typedef {import("./types").TsLibType} TsLibType
 * @typedef {import("./types").HttpClientLibTypeV2} HttpClientLibTypeV2
 * @typedef {import("./types").TsSwaggerGenerateOptions} TsSwaggerGenerateOptions
 * @typedef {import("./types").TsSwaggerGenerateOptionsV2} TsSwaggerGenerateOptionsV2
 */

const assert = require("assert");
const chalk = require("chalk");
const { program } = require("commander");
const { generateTs } = require("./swaggerTs");
const { generateTsV2 } = require("./swaggerTsV2");
const { generateCs } = require("./swaggerCs");
const { generateAndroid } = require("./swaggerAndroid");

const { OPEN_TMP_FOLDER } = require("./env");

const updateNotifier = require("update-notifier");
const pkg = require("../package.json");
const { validate } = require("./validateSwagger");

const notifier = updateNotifier({ pkg, });

notifier.notify({
	boxenOptions: {
		padding: 1,
		margin: 1,
		align: "center",
		borderColor: "red",
	},
});

if (notifier.update) {
	console.error(chalk.bgRed(`Please run \`ncu -u ${notifier.update.name}\` to update to the latest version: ${notifier.update.latest}`));
	process.exit(1);
}
console.log("fht-swagger-cli: " + pkg.version);

(async () => {
	program
		.version(pkg.version)
		.command("ts <tsLibType>")
		.description("生成单个api.ts, 包含所有的api")
		.requiredOption("-e --endpointUrl <endpointUrl>", "swagger doc的路径", /^https?:\/\/.+/i)
		.option("-o --outputPath <outputPath>", "生成的api ts文件路径", "./api.ts")
		.option("-c --classValidator", "javascript annotation验证，使用https://github.com/typestack/class-validator")
		.option("-r --ruleValidator", "async-validator的rules规则, 使用https://github.com/yiminghe/async-validator")
		.option(
			"-s --customValidator",
			"是否生成customValidator,如sensitive-world, html-length, 等规则，api.ts的用户需要提供这些规则与 `./customValidators里面`",
		)
		.option("-m --debugModels <debugModelsPath>")
		.option("-a --debugOperations <debugOperationsPath>")
		.option(
			"-p --apiBasePathEnvPrefix <apiBasePathEnvPrefix>",
			"获取globalThis下面的对应属性值作为api的base path",
			"API_BASE_PATH",
		)
		.action(
			/**
			 * @param {TsLibType} tsLibType
			 * @param {*} cmdObj
			 */
			async function (tsLibType, cmdObj) {
				/** @type {string} */
				const endpointUrl = cmdObj.endpointUrl;
				/** @type {string} */
				const outputPath = cmdObj.outputPath;
				const classValidator = !!cmdObj.classValidator;
				const ruleValidator = !!cmdObj.ruleValidator;
				const customValidator = !!cmdObj.customValidator;
				/** @type {string|undefined} */
				const debugModels = cmdObj.debugModels;
				/** @type {string|undefined} */
				const debugOperations = cmdObj.debugOperations;
				const apiBasePathEnvPrefix = cmdObj.apiBasePathEnvPrefix;
				await validate(endpointUrl);
				await generateTs({
					tsLibType,
					endpointUrl,
					outputPath,
					debugModels,
					debugOperations,
					classValidator,
					ruleValidator,
					customValidator,
					apiBasePathEnvPrefix,
				});
			},
		);

	/** @type {HttpClientLibTypeV2[]} */
	const httpClientLibOptions = ["jquery", "axios", "fetch", "taro"];

	/**
	 * @param {string} lib 
	 */
	function validateHttpClientLib(lib) {
		assert(lib, "lib falsy");
		/** @type {HttpClientLibTypeV2} */
		// @ts-ignore
		const copy = lib;
		if (httpClientLibOptions.includes(copy)) return copy;
		return;
	}

	program
		.command("ts2")
		.description("每个API单独文件")
		.requiredOption("-l --httpClientLib <httpClientLib>", `httpClientLib, one of: ${httpClientLibOptions.join(", ")}`, validateHttpClientLib)
		.requiredOption("-e --endpointUrl <endpointUrl>", "swagger doc的路径", /^https?:\/\/.+/i)
		.option("-o --outputDir <outputDir>", "生成的api的文件夹路径", "./apis")
		.option("--classValidator", "javascript annotation验证，使用https://github.com/typestack/class-validator")
		.option("--ruleValidator", "async-validator的rules规则, 使用https://github.com/yiminghe/async-validator")
		.option(
			"--customValidator",
			"是否生成customValidator,如sensitive-world, html-length, 等规则，api.ts的用户需要提供这些规则与 `./customValidators里面`",
		)
		.option(
			"--customValidatorRelativeLevel <customValidatorRelativeLevel>",
			"customValidators.ts相对于api/api.ts的文件路径，默认2级，也就是从api/api.ts的父文件夹的父文件夹与customValidators.ts同级",
		)
		.option("--convertFromEnum", "generate method that converts from enum to number")
		.option("--convertToEnum", "generate method that converts from number to enum")
		.option("--describeEnum", "generate method that converts enum into its description")
		.option("-m --debugModels <debugModelsPath>")
		.option("-a --debugOperations <debugOperationsPath>")
		.option(
			"-p --apiBasePathEnvPrefix <apiBasePathEnvPrefix>",
			"获取globalThis下面的对应属性值作为api的base path",
			"API_BASE_PATH",
		)
		.action(
			/**
			 * @param {TsSwaggerGenerateOptionsV2} cmdObj
			 */
			async function (cmdObj) {
				/** @type {string} */
				const endpointUrl = cmdObj.endpointUrl;
				assert(typeof endpointUrl === "string", "endpointUrl must be string");
				await validate(endpointUrl);
				/** @type {TsSwaggerGenerateOptionsV2} */
				const config = {
					httpClientLib: cmdObj.httpClientLib,
					endpointUrl: cmdObj.endpointUrl,
					outputDir: cmdObj.outputDir,
					debugModels: cmdObj.debugModels,
					debugOperations: cmdObj.debugOperations,
					classValidator: cmdObj.classValidator,
					ruleValidator: cmdObj.ruleValidator,
					customValidator: cmdObj.customValidator,
					customValidatorRelativeLevel: Number(cmdObj.customValidatorRelativeLevel) || 2,
					convertToEnum: cmdObj.convertToEnum,
					convertFromEnum: cmdObj.convertFromEnum,
					describeEnum: cmdObj.describeEnum,
					apiBasePathEnvPrefix: cmdObj.apiBasePathEnvPrefix,
				}
				await generateTsV2(config);
				console.log("DONE generateTsV2 with config:", config);
				process.exit(0);
			},
		);
	program
		.command("android")
		.requiredOption("-e --endpointUrl <endpointUrl>", "swagger doc的路径", /^https?:\/\/.+/i)
		.option("-m --debugModels <debugModelsPath>")
		.option("-a --debugOperations <debugOperationsPath>")
		.option("-o --outputDir <outputDir>", "生成的源文件的输出文件夹，不指定就不生成")
		.action(async function (cmdObj) {
			/** @type {string} */
			const endpointUrl = cmdObj.endpointUrl;
			/** @type {string|undefined} */
			const outputDir = cmdObj.outputDir;
			/** @type {string|undefined} */
			const debugModels = cmdObj.debugModels;
			/** @type {string|undefined} */
			const debugOperations = cmdObj.debugOperations;
			await validate(endpointUrl);
			await generateAndroid({
				endpointUrl,
				outputDir,
				debugModels,
				debugOperations,
			});
		});

	program
		.command("cs")
		.requiredOption("-e --endpointUrl <endpointUrl>", "swagger doc的路径", /^https?:\/\/.+/i)
		.option(
			"-p --packageOutputDir <packageOutputDir>",
			"生成的nuget package包存入的文件夹",
			"\\\\192.168.1.228\\huohuoshared\\NugetPackages",
		)
		.option("-m --debugModels <debugModelsPath>")
		.option("-a --debugOperations <debugOperationsPath>")
		.option("-o --outputDir <outputDir>", "生成的源文件的输出文件夹，不指定就不生成")
		.option("-n --packageName <packageName>", "package Name", "IO.Swagger")
		.action(async function (cmdObj) {
			/** @type {string} */
			const endpointUrl = cmdObj.endpointUrl;
			/** @type {string} */
			const packageOutputDir = cmdObj.packageOutputDir;
			/** @type {string} */
			const packageName = cmdObj.packageName;
			/** @type {string|undefined} */
			const outputDir = cmdObj.outputDir;
			/** @type {string|undefined} */
			const debugModels = cmdObj.debugModels;
			/** @type {string|undefined} */
			const debugOperations = cmdObj.debugOperations;
			await validate(endpointUrl);
			await generateCs({
				endpointUrl,
				packageOutputDir,
				packageName,
				outputDir,
				debugModels,
				debugOperations,
			});
		});

	program.parse(process.argv);
})();
