// @ts-check

/**
 * @typedef {import("./types").AndroidSwaggerGenerateOptions} AndroidSwaggerGenerateOptions
 */

const assert = require("assert");
const path = require("path");
const fs = require("fs");
const chalk = require("chalk");
const shelljs = require("shelljs");
const execa = require("execa");
const tmp = require("tmp");

const { swaggerCodegen2_2_3: swaggerJarPath } = require("fht-swagger-jar");
assert(fs.existsSync(swaggerJarPath), "jarPath not found");

shelljs.config.fatal = true;

/**
 * @param {AndroidSwaggerGenerateOptions} opts
 */
async function generateAndroid(opts) {
	console.log("Swagger codegen endpoint:", opts.endpointUrl);

	const tmpobj = tmp.dirSync();
	const tmpDir = tmpobj.name;

	if (opts.debugModels) {
		console.log(chalk.bgYellow("Debug Models"));
		const cmd = `java -DdebugModels=true -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l android -o "${tmpDir}" > ${opts.debugModels}`;
		console.log("EXECUTING: ", cmd);
		const res = shelljs.exec(cmd);
		if (res.code) {
			console.error(chalk.bgRed("Swagger codegen failed"));
			console.error(chalk.red(res.stderr));
			return process.exit(1);
		}
		return;
	}
	if (opts.debugOperations) {
		console.log(chalk.bgYellow("Debug Operations"));
		const cmd = `java -DdebugOperations=true -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l android -o "${tmpDir}" > ${opts.debugOperations}`;
		const res = shelljs.exec(cmd);
		if (res.code) {
			console.error(chalk.bgRed("Swagger codegen failed"));
			console.error(chalk.red(res.stderr));
			return process.exit(1);
		}
		return;
	}

	// java -jar swagger-codegen-cli-2.2.3.jar generate -i %SWAGGERENDPOINT%
	// -l android -o ./swaggerCsharp -t fhtAndroidTemplates
	const templateDir = path.join(__dirname, "../fhtAndroidTemplates");
	const configDir = path.join(__dirname, "../swagger_android_volley.config");
	assert(fs.existsSync(templateDir), `templateDir ${templateDir} does not exist`);
	assert(fs.existsSync(configDir), `configDir ${configDir} does not exist`);
	const cmd = `java -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l android -t ${templateDir} -c ${configDir} -o "${tmpDir}"`;
	const res = shelljs.exec(cmd);
	if (res.code) {
		console.error(chalk.bgRed("Swagger codegen failed"));
		console.error(chalk.red(res.stderr));
		return process.exit(1);
	}
	if (opts.outputDir) {
		if (!fs.existsSync(opts.outputDir)) {
			shelljs.mkdir("-p", opts.outputDir);
		}
		const swaggerSrcDir = path.join(tmpDir, "./src/main/java/com/fht360/swagger");
		assert(fs.existsSync(swaggerSrcDir), `swaggerSrcDir ${swaggerSrcDir} does not exist`);
		for (const folder of ["models", "apis", "invoker"]) {
			const fullDir = path.join(swaggerSrcDir, folder);
			shelljs.cp("-rf", fullDir, opts.outputDir);
		}
		const swaggerClientSrcDir = path.join(tmpDir, "./src/main/java/io/swagger/client");
		assert(fs.existsSync(swaggerClientSrcDir), `swaggerClientSrcDir ${swaggerClientSrcDir} does not exist`);
		for (const folder of ["auth", "request"]) {
			const fullDir = path.join(swaggerClientSrcDir, folder);
			shelljs.cp("-rf", fullDir, path.join(opts.outputDir, "invoker"));
		}
		console.log(chalk.bgGreen("生成代码到: " + opts.outputDir));
	}
	tmpobj.removeCallback();
}

module.exports = {
	generateAndroid,
};
