// @ts-check

/**
 * @typedef {import("./types").CsSwaggerGenerateOptions} CsSwaggerGenerateOptions
 */

/** @type {import("assert")} */
const assert = require("assert");
const path = require("path");
const fs = require("fs");
const chalk = require("chalk");
const shelljs = require("shelljs");
const execa = require("execa");
const tmp = require("tmp");
const _ = require("lodash");
// @ts-ignore
const { swaggerCodegen2_2_3: swaggerJarPath } = require("fht-swagger-jar");
// @ts-ignore
const nugetExePath = require("fht-nuget-exe");
const { locateMsBuildExe, makeNugetVersionBuildNo } = require("./utils");
const globby = require("globby");

shelljs.config.fatal = true;

assert(fs.existsSync(swaggerJarPath), "jarPath not found");

/**
 * @param {CsSwaggerGenerateOptions} opts
 */
async function generateCs(opts) {
	assert(fs.existsSync(nugetExePath), "nuget.exe not found");
	console.log("Swagger codegen endpoint:", opts.endpointUrl);

	const tmpobj = tmp.dirSync();
	const msbuildExe = await locateMsBuildExe();
	const tmpDir = tmpobj.name;

	if (opts.debugModels) {
		console.log(chalk.bgYellow("Debug Models"));
		//java -DdebugModels -jar swagger-codegen-cli-2.2.3.jar generate -i http://192.168.1.223:8000/swagger/docs/v1 -c swagger_typescript_jquery.config -l typescript-jquery >| debugmodels_ts_jquery_promise_codegen.json
		const cmd = `java -DdebugModels=true -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l csharp -o "${tmpDir}" > ${opts.debugModels}`;
		console.log("EXECUTING: ", cmd);
		const res = shelljs.exec(cmd);
		if (res.code) {
			console.error(chalk.bgRed("Swagger codegen failed"));
			console.error(chalk.red(res.stderr));
			return process.exit(1);
		}
		return;
	}
	if (opts.debugOperations) {
		console.log(chalk.bgYellow("Debug Operations"));
		const cmd = `java -DdebugOperations=true -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l csharp -o "${tmpDir}" > ${opts.debugOperations}`;
		const res = shelljs.exec(cmd);
		if (res.code) {
			console.error(chalk.bgRed("Swagger codegen failed"));
			console.error(chalk.red(res.stderr));
			return process.exit(1);
		}
		return;
	}

	// java -jar swagger-codegen-cli-2.2.3.jar generate -i %SWAGGERENDPOINT%
	// -l csharp -o ./swaggerCsharp -t fhtCsharpTemplates
	const templateDir = path.join(__dirname, "../fhtCsharpTemplates");
	const packageName = opts.packageName;
	const cmd = `java -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l csharp -t ${templateDir} -DpackageName=${packageName} -o "${tmpDir}"`;
	const res = shelljs.exec(cmd);
	if (res.code) {
		console.error(chalk.bgRed("Swagger codegen failed"));
		console.error(chalk.red(res.stderr));
		return process.exit(1);
	}
	//remove a few .net framework relic
	shelljs.rm("-rf", tmpDir + `/src/${opts.packageName}/Properties`);
	shelljs.rm(tmpDir + `/src/${opts.packageName}/*.nuspec`);

	// const tmpDir = "C:\\Users\\ADMINI~1\\AppData\\Local\\Temp\\tmp-1496ECEPA4jEs7tT";
	if (opts.outputDir) {
		if (!fs.existsSync(opts.outputDir)) {
			shelljs.mkdir("-p", opts.outputDir);
		}
		shelljs.cp("-f", tmpDir + "/*.sln", opts.outputDir);
		shelljs.cp("-rf", tmpDir + "/src", opts.outputDir);
		console.log(chalk.bgGreen("生成代码到: " + opts.outputDir));
	}

	const modelFiles = await globby("./src/**/Model/**/*.cs", {
		cwd: tmpDir,
		absolute: true,
	});

	//a hack to work around a limitation
	shelljs.sed("-i", /public\s+(\w+)\?(.*x-isnullable=false)/, "public $1 $2", modelFiles);

	const slnPath = path.join(tmpDir, `${packageName}.sln`);
	const projectPath = path.join(tmpDir, `src/${packageName}/${packageName}.csproj`);
	console.log(chalk.bgGreen("Nuget restoring:"), slnPath);
	const restoreResult = await execa(nugetExePath, ["restore", slnPath], {
		shell: true,
	});

	if (restoreResult.exitCode) {
		console.log(chalk.bgRed("Failed to nuget restore: " + projectPath));
		throw new Error(restoreResult.stderr);
	}

	console.log(chalk.bgGreen("MsBuilding:"), msbuildExe, projectPath);
	const msBuildResult = await execa(
		`"${msbuildExe}"`,
		[projectPath, "/t:Build", "/m:4", "/p:BuildInParallel=true", "/p:Configuration=Release"],
		{
			reject: false,
		},
	);

	if (msBuildResult.exitCode || msBuildResult.failed) {
		console.log(chalk.bgRed("Failed to msbuild: " + projectPath));
		const errors = _.chain(msBuildResult.stdout.split("\n")).filter(l => !!l && l.includes(": error "))
			.map(l => l.trim()).value();
		const errorMsg = msBuildResult.stderr + "\r\n" + errors.join("\r\n");
		console.error(chalk.red(errorMsg));
		throw new Error("Failed to msbuild: " + projectPath);
	} else {
		console.log(chalk.bgGreen(msBuildResult.command), "SUCCESSFUL");
		if (msBuildResult.stderr) {
			console.error(chalk.bgYellow("But with some error:"), chalk.bgRed(msBuildResult.stderr));
		}
	}

	if (opts.packageOutputDir) {
		// %NUGET% pack FHT.Core.Configurations.csproj -IncludeReferencedProjects -Version 1.0.%version% -Verbosity detailed -Prop Configuration=Release
		console.log(chalk.bgGreen("Nuget packing:"), projectPath, "to", opts.packageOutputDir);
		assert(fs.existsSync(nugetExePath), `nugetExe: ${nugetExePath} does not exist.`);
		assert(fs.existsSync(projectPath), `projectPath: ${projectPath} does not exist.`);
		assert(fs.existsSync(tmpDir), `tmpDir: ${tmpDir} does not exist.`);
		const versionNo = makeNugetVersionBuildNo();
		// const outputDirectory = "\\\\192.168.1.228\\huohuoshared\\NugetPackages";
		const packResult = await execa(
			nugetExePath,
			[
				"pack",
				projectPath,
				"-IncludeReferencedProjects",
				"-ExcludeEmptyDirectories",
				// `-MSBuildPath "${msbuildExe}"`,
				`-Version 1.0.${versionNo}`,
				"-Prop Configuration=Release",
				`-OutputDirectory ${tmpDir}`,
			],
			{
				shell: true,
			},
		);
		if (packResult.exitCode) {
			console.log(chalk.bgRed("Failed to nuget pack: " + projectPath));
			throw new Error(packResult.stdout + "\r\n" + packResult.stderr);
		} else {
			console.log(chalk.bgGreen(packResult.command));
			console.log(chalk.green(packResult.stdout));
		}

		const packageOutputDir = path.normalize(opts.packageOutputDir);
		if (!fs.existsSync(packageOutputDir)) {
			shelljs.mkdir("-p", packageOutputDir);
		}
		console.log("Copy nupkg to", packageOutputDir);
		shelljs.mv(path.join(tmpDir, "*.nupkg"), path.join(packageOutputDir));
		console.log(chalk.bgGreen("成功生成nuget包到: " + path.join(packageOutputDir)));
	}
	tmpobj.removeCallback();
}

module.exports = {
	generateCs,
};
