//@ts-check
const path = require("path");
const fs = require("fs");
const chalk = require("chalk");
const shelljs = require("shelljs");
const tmp = require("tmp");
const assert = require("assert");
const { swaggerCodegen2_2_3: swaggerJarPath } = require("fht-swagger-jar");

const { fixUseBeforeDeclaration } = require("./fixUseBeforeDeclaration");

assert(fs.existsSync(swaggerJarPath), "jarPath not found");

/**
 * @typedef {import("./types").TsLibType} TsLibType
 * @typedef {import("./types").TsSwaggerGenerateOptions} TsSwaggerGenerateOptions
 */
const resolveInTemplate = /**
     * @param {string} p 
     */
	(p) => path.resolve(__dirname, "../fhtTypescriptTemplates", p);

/**
 * @param {TsSwaggerGenerateOptions} opts
 */
async function generateTs(opts) {
	/** @type {{[k in TsLibType]: {lang: string, templatePath: string, configPath: string }}} */
	const tsLibTypeMap = {
		fetch: {
			lang: "typescript-fetch",
			templatePath: resolveInTemplate("fetch"),
			configPath: resolveInTemplate("swagger_typescript_fetch.config"),
		},
		jquery: {
			lang: "typescript-jquery",
			templatePath: resolveInTemplate("jquery"),
			configPath: resolveInTemplate("swagger_typescript_jquery.config"),
		},
		request: {
			lang: "typescript-node",
			templatePath: resolveInTemplate("request"),
			configPath: resolveInTemplate("swagger_typescript_node.config"),
		},
		taro: {
			lang: "typescript-node",
			templatePath: resolveInTemplate("taro"),
			configPath: resolveInTemplate("swagger_typescript_taro.config"),
		},
	};
	const config = tsLibTypeMap[opts.tsLibType];
	if (!config) {
		console.error(
			"tsLibType can only be one of: " + Object.keys(tsLibTypeMap).join(", ") + ", but is: " + opts.tsLibType,
		);
		return process.exit(1);
	}

	const tmpobj = tmp.dirSync();
	const tmpDir = tmpobj.name;
	if (opts.debugModels) {
		console.log(chalk.bgYellow("Debug Models"));
		//java -DdebugModels -jar swagger-codegen-cli-2.2.3.jar generate -i http://192.168.1.223:8000/swagger/docs/v1 -c swagger_typescript_jquery.config -l typescript-jquery >| debugmodels_ts_jquery_promise_codegen.json
		const cmd = `java -DdebugModels=true -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l "${config.lang}" -o "${tmpDir}" -c "${config.configPath}" > ${opts.debugModels}`;
		console.log("RUN:", cmd);
		const res = shelljs.exec(cmd);
		if (res.code) {
			console.error(chalk.bgRed("Swagger codegen failed"));
			console.error(chalk.red(res.stderr));
			return process.exit(1);
		}
		return;
	}
	if (opts.debugOperations) {
		console.log(chalk.bgYellow("Debug Operations"));
		const cmd = `java -DdebugOperations=true -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l "${config.lang}" -o "${tmpDir}" -c "${config.configPath}" > ${opts.debugOperations}`;
		console.log("RUN:", cmd);
		const res = shelljs.exec(cmd);
		if (res.code) {
			console.error(chalk.bgRed("Swagger codegen failed"));
			console.error(chalk.red(res.stderr));
			return process.exit(1);
		}
		return;
	}

	let additional = "";
	if (opts.classValidator) {
		additional += " --additional-properties supportsClassValidator=true ";
	}
	if (opts.ruleValidator) {
		additional += " --additional-properties supportsRuleValidator=true ";
	}
	if (opts.customValidator) {
		additional += " --additional-properties supportsCustomValidator=true ";
	}
	if (opts.apiBasePathEnvPrefix) {
		additional += ` --additional-properties apiBasePathEnvPrefix=${opts.apiBasePathEnvPrefix} `;
	}

	const cmd = `java -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l "${config.lang}" -t "${config.templatePath}" -o "${tmpDir}" -c "${config.configPath}" ${additional}`;
	console.log("RUN:", cmd);
	const res = shelljs.exec(cmd);
	if (res.code) {
		console.error(chalk.bgRed("Swagger codegen failed"));
		console.error(chalk.red(res.stderr));
		return process.exit(1);
	}
	const outputDir = path.dirname(opts.outputPath);
	if (!fs.existsSync(outputDir)) {
		shelljs.mkdir("-p", outputDir);
	}
	shelljs.mv(path.join(tmpDir, "api.ts"), opts.outputPath);
	await fixUseBeforeDeclaration(opts.outputPath);
	tmpobj.removeCallback();
}

module.exports = {
	generateTs,
};
