//@ts-check

const path = require("path");
const fs = require("fs-extra");
const _ = require("lodash");
const chalk = require("chalk");
const shelljs = require("shelljs");
const tmp = require("tmp");
const assert = require("assert");
const { swaggerCodegen2_4_22: swaggerJarPath } = require("fht-swagger-jar");

const { fixApiModelImportsV2 } = require("./fixApiModelImportsV2");
const { OPEN_TMP_FOLDER } = require("./env");
const { truncateSync } = require("fs");

/**
 * @typedef {import("./types").TsSwaggerGenerateOptionsV2} TsSwaggerGenerateOptionsV2
 */
const resolveInTemplate = /**
     * @param {string} p 
     */
	(p) => path.resolve(__dirname, "../fhtTypescriptTemplates", p);

/**
 * @param {TsSwaggerGenerateOptionsV2} opts
 */
async function generateTsV2(opts) {
	const config = {
		lang: "typescript-jquery",
		templatePath: resolveInTemplate("v2"),
		configPath: resolveInTemplate("swagger_typescript_jquery_v2.json"),
	}

	const tmpobj = tmp.dirSync();
	const tmpDir = tmpobj.name;
	if (opts.debugModels) {
		console.log(chalk.bgYellow("Debug Models"));
		//java -DdebugModels -jar swagger-codegen-cli-2.2.3.jar generate -i http://192.168.1.223:8000/swagger/docs/v1 -c swagger_typescript_jquery.config -l typescript-jquery >| debugmodels_ts_jquery_promise_codegen.json
		const cmd = `java -DdebugModels=true -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l "${config.lang}" -o "${tmpDir}" -c "${config.configPath}" > ${opts.debugModels}`;
		console.log("RUN:", cmd);
		const res = shelljs.exec(cmd);
		if (res.code) {
			console.error(chalk.bgRed("Swagger codegen failed"));
			console.error(chalk.red(res.stderr));
			return process.exit(1);
		}
		return;
	}
	if (opts.debugOperations) {
		console.log(chalk.bgYellow("Debug Operations"));
		const cmd = `java -DdebugOperations=true -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l "${config.lang}" -o "${tmpDir}" -c "${config.configPath}" > ${opts.debugOperations}`;
		console.log("RUN:", cmd);
		const res = shelljs.exec(cmd);
		if (res.code) {
			console.error(chalk.bgRed("Swagger codegen failed"));
			console.error(chalk.red(res.stderr));
			return process.exit(1);
		}
		return;
	}

	assert(opts.httpClientLib, "opts.httpClientLib falsy");

	let additional = "";
	additional += ` --additional-properties httpClientLib${_.capitalize(opts.httpClientLib)} `;
	if (opts.classValidator) {
		additional += " --additional-properties supportsClassValidator=true ";
	}
	if (opts.ruleValidator) {
		additional += " --additional-properties supportsRuleValidator=true ";
	}
	if (opts.customValidator) {
		additional += " --additional-properties supportsCustomValidator=true ";
	}
	if (opts.customValidatorRelativeLevel) {
		const relativePath = _.repeat("../", opts.customValidatorRelativeLevel)
		additional += ` --additional-properties customValidatorRelativePath=${relativePath} `;
	}
	if (opts.convertFromEnum) {
		additional += " --additional-properties supportsConvertFromEnum=true ";
	}
	if (opts.convertToEnum) {
		additional += " --additional-properties supportsConvertToEnum=true ";
	}
	if (opts.describeEnum) {
		additional += " --additional-properties supportsDescribeEnum=true ";
	}
	if (opts.apiBasePathEnvPrefix) {
		additional += ` --additional-properties apiBasePathEnvPrefix=${opts.apiBasePathEnvPrefix} `;
	}

	const cmd = `java -jar "${swaggerJarPath}" generate -i "${opts.endpointUrl}" -l "${config.lang}" -t "${config.templatePath}" -o "${tmpDir}" -c "${config.configPath}" ${additional}`;
	console.log("RUN:", cmd);
	const res = shelljs.exec(cmd);
	if (res.code) {
		console.error(chalk.bgRed("Swagger codegen failed"));
		console.error(chalk.red(res.stderr));
		return process.exit(1);
	}
	const outputDir = opts.outputDir;
	assert(outputDir, "outputDir falsy")
	await fs.ensureDir(outputDir);
	await fs.emptyDir(outputDir);
	tmpobj.removeCallback();

	/** @type {(r:string)=> string} */
	const resolve = r => path.join(tmpDir, r);

	if (opts.customValidator) {
		await fs.outputFile(resolve("../customValidators.ts"), `
export const sensitiveWord: any = "dummy";
export const htmlTextLength: any = "dummy";
export const elementCount: any = "dummy";
export const birthYear: any = "dummy";
export const enforceTrue: any = "dummy";		
		`, { encoding: "utf-8" });
	}

	if (OPEN_TMP_FOLDER) {
		require("open")(tmpDir, {
			app: {
				name: "code",
			}
		});
	}
	await fixApiModelImportsV2(tmpDir);
	console.log("Copying ts files from", tmpDir, "to", opts.outputDir);
	shelljs.cp("-r", resolve("./api"), opts.outputDir);
	shelljs.cp("-r", resolve("./model"), opts.outputDir);
	console.log("SWAGGER CODEGEN SUCCESSFUL");
}

module.exports = {
	generateTsV2,
};

if (require.main === module) {
	(async () => {
		const outputDir = process.env.OUTPUT_DIR || "./cli/outputs/v2";
		await generateTsV2({
			endpointUrl: "http://192.168.1.223:8000/swagger/docs/v1",
			outputDir,
			// classValidator: true,
			// ruleValidator: true,
			// customValidator: true,
			customValidatorRelativeLevel: 2,
			httpClientLib: "axios",
			apiBasePathEnvPrefix: "API_BASE_PATH",
		})
		console.log("DONE generating jquery example v2");
	})();
}