//@ts-check
const { generateTs } = require("./swaggerTs");
const assert = require("assert");
const path = require("path");
const ts = require("typescript");
const chalk = require("chalk");
const util = require("util");
const { validate } = require("./validateSwagger");

/**
 * @typedef {import("./types").TsLibType} TsLibType
 * @typedef {import("./outputs/output_api_request").TestSwaggerInputModel} TestSwaggerInputModel
 * @typedef {keyof TestSwaggerInputModel} TestSwaggerInputModelKeys
 * @typedef {import("./outputs/output_api_request").TestSwaggerViewModel} TestSwaggerViewModel
 * @typedef {keyof TestSwaggerViewModel} TestSwaggerViewModelKeys
 */

const resolve = (p) => path.resolve(__dirname, p);

//see https://github.com/Microsoft/TypeScript/wiki/Using-the-Compiler-API
//also https://github.com/Microsoft/TypeScript/issues?utf8=%E2%9C%93&q=is%3Amerged%20label%3A%22Domain%3A%20Transforms%22%20
//also https://github.com/microsoft/TypeScript/pull/13940
//also https://ts-ast-viewer.com/
//also https://blog.scottlogic.com/2017/05/02/typescript-compiler-api-revisited.html

(async () => {
	// const swaggerDocUrl = "http://localhost:2000/swagger/docs/v1";
	// const swaggerDocUrl = "http://www.mobiledevelopmentfht360web.com/swagger/docs/v1";
	const swaggerDocUrl = "http://192.168.1.223:8000/swagger/docs/v1";
	/** @type {TsLibType[]} */
	// const tsLibTypes = ["fetch", "jquery", "request"];
	const tsLibTypes = ["jquery"];
	for (const tsLibType of tsLibTypes) {
		console.log(chalk.bgYellow(`=== swagger ts ${tsLibType} === `));
		const outputTsPath = resolve(`./outputs/output_api_${tsLibType}.ts`);
		await validate(swaggerDocUrl);
		await generateTs({
			tsLibType,
			endpointUrl: swaggerDocUrl,
			outputPath: outputTsPath,
			ruleValidator: true,
			customValidator: true,
			apiBasePathEnvPrefix: "API_ROOT_PATH",
		});
		const program = ts.createProgram([outputTsPath], {});
		const checker = program.getTypeChecker();
		const printer = ts.createPrinter({ newLine: ts.NewLineKind.LineFeed });
		const diagnostics = program.getSyntacticDiagnostics();
		console.log("===========================");
		if (diagnostics.length) {
			console.log(chalk.bgRed(`${diagnostics.length} errors found`));
			for (const d of diagnostics) {
				console.log(chalk.red(`TS${d.code}`), d.messageText, `check file: ${outputTsPath}`);
			}
		}
		console.log("===========================");
		for (const sc of program.getSourceFiles()) {
			if (sc.isDeclarationFile) continue;
			ts.forEachChild(sc, (node) => {
				if (ts.isClassDeclaration(node) || ts.isInterfaceDeclaration(node)) {
					const className = node.name.escapedText.toString();
					if (className === "TestSwaggerInputModel" || className === "TestSwaggerViewModel") {
						if (className === "TestSwaggerViewModel") {
							/** @type {string[]} */
							const baseNames = [];
							if (node.heritageClauses) {
								for (const hc of node.heritageClauses) {
									if (hc.token === ts.SyntaxKind.ExtendsKeyword) {
										for (const t of hc.types) {
											const baseName = printer.printNode(ts.EmitHint.Unspecified, t, sc);
											baseNames.push(baseName);
										}
									}
								}
							}
							assert(
								baseNames.includes("TestSwaggerViewModelBase"),
								"must inherit from TestSwaggerViewModelBase",
							);
						}
						for (const member of node.members) {
							// console.log(member.name.getText(), util.inspect(memberType, { depth: 1 }));
							// console.log(member.name.getText(), checker.typeToString(memberType));
							// console.log(member.getText());
							let memberName = member.name.getText();
							const isStatic =
								member.modifiers &&
								member.modifiers.some((m) => m.kind === ts.SyntaxKind.StaticKeyword);
							if (isStatic) continue;
							assert(memberName.startsWith("'") && memberName.endsWith("'"), memberName);
							if (ts.isPropertyDeclaration(member) || ts.isPropertySignature(member)) {
								const isRequired = !member.questionToken;
								const memberType = checker.getTypeAtLocation(member.name);
								if (className === "TestSwaggerInputModel") {
									/** @type {TestSwaggerInputModelKeys} */
									// @ts-ignore
									const propName = memberName.substr(1, memberName.length - 2);
									if (propName === "RequiredInt") {
										assert(isRequired, `${propName} should required`);
									} else if (propName === "NonRequiredInt") {
										assert(
											isRequired,
											`${propName} should also be required because it is non-nullable value type`,
										);
									} else if (propName === "NullableInt") {
										assert(!isRequired, `${propName} should not be required`);
									} else if (propName === "RequiredString") {
										assert(isRequired, `${propName} should be required`);
									} else if (propName === "NonRequiredString") {
										assert(!isRequired, `${propName} should not be required`);
									} else if (propName === "Gender") {
										assert(isRequired, `${propName} should be required`);
									} else if (propName === "NullableGender") {
										assert(!isRequired, `${propName} should not be required`);
									} else if (propName === "Genders") {
										const typeString = checker.typeToString(memberType);
										assert(typeString.includes("[]"), propName + " must be array type");
									} else if (propName === "RequiredChild") {
										assert(isRequired, `${propName} should not be required`);
									} else if (propName === "Child") {
										assert(!isRequired, `${propName} should not be required`);
									} else if (propName === "RequiredChildren") {
										assert(isRequired, `${propName} should not be required`);
									} else if (propName === "Children") {
										assert(!isRequired, `${propName} should not be required`);
										const typeString = checker.typeToString(memberType);
										assert(typeString.includes("[]"), propName + " must be array type");
									} else if (propName === "Inner") {
										console.log("skip", propName);
									} else {
										throw new Error("add support for " + propName);
									}
								} else {
									/** @type {TestSwaggerViewModelKeys} */
									// @ts-ignore
									const propName = memberName.substr(1, memberName.length - 2);
									if (propName === "NonNullableInt") {
										assert(isRequired, `${propName} should be required`);
									} else if (propName === "NullableInt") {
										assert(!isRequired, `${propName} should not be required`);
									} else if (propName === "NullableGender") {
										assert(!isRequired, `${propName} should not be required`);
									} else if (propName === "NonNullableGender") {
										assert(isRequired, `${propName} should be required`);
									} else if (propName === "Genders") {
										assert(
											isRequired,
											`${propName}, a collection prop in view model, should be required`,
										);
										const typeString = checker.typeToString(memberType);
										assert(typeString.includes("[]"), propName + " must be array type");
									} else if (propName === "Child") {
										assert(isRequired, `${propName}, a ref prop in view model, should be required`);
									} else if (propName === "Children") {
										assert(
											isRequired,
											`${propName}, a collection prop in view model, should be required`,
										);
										const typeString = checker.typeToString(memberType);
										assert(typeString.includes("[]"), propName + " must be array type");
									} else if (propName === "String") {
										assert(
											isRequired,
											`${propName}, a string prop in view model, should be required`,
										);
									} else if (propName === "Message") {
									} else {
										throw new Error("pls add support for " + propName);
									}
								}
							}
						}
					}
				}
			});
		}
	}
})();
