//@ts-check

const assert = require("assert");
const axios = require("axios").default;
const chalk = require("chalk");
const jp = require("jsonpath");

const X_TYPE_INHERITS = 'x-inherits';
const X_TYPE_INHERITS_NAME = 'x-inheritsName';
const X_TYPE_INHERITS_FULLNAME = 'x-inheritsFullName';
const X_PROP_IS_INHERITED = 'x-isinherited';
const X_ENUM_TYPE_NAME = 'x-enumTypeName';

const ENV_SKIP_SWAGGER_CHECK_KEY = "SKIP_SWAGGER_CHECK";

/**
 * @param {{[k:string]:SwaggerDefinition}} definitions
 * @returns {string[]}
 */
function validateInheritance(definitions) {
	assert(definitions, "definitions falsy");
	/** @type {string[]} */
	const violations = [];
	for (const [name, def] of Object.entries(definitions)) {
		if (!def.properties) continue;
		const hasIsInheritedProp = Object.values(def.properties).some(p => p[X_PROP_IS_INHERITED])
		const isInheritedType = def[X_TYPE_INHERITS]
		if (hasIsInheritedProp && !isInheritedType) {
			violations.push(`${name}'s base class should be annotated with IsPolymorphicBaseClass`);
		}
	}
	return violations;
}

/**
 * @param {{[k:string]:SwaggerDefinition}} definitions
 * @returns {string[]}
 */
function validateEnumDefinition(definitions) {
	assert(definitions, "definitions falsy");
	/** @type {string[]} */
	const violations = [];
	for (const [name, def] of Object.entries(definitions)) {
		if (!def.enum) continue;
		const enumName = def["x-enumTypeName"];
		try {
			assert(enumName, `${enumName} gotta have x-enumTypeName`);
			assert(def["x-has-enum-pairs"], `${enumName} does not have x-has-enum-pairs`);
			assert(def["x-enum-pairs"], `${enumName} does not have x-enum-pairs`);
		} catch (err) {
			/** @type  {Error} */
			const error = err;
			const errMsg = error.message;
			violations.push(errMsg);
		}
	}
	return violations;
}

/**
 * @param {{ [k: string]: SwaggerDefinition; }} definitions
 * @returns {string[]}
 */
function validateEnumInAllPositions(definitions) {
	const enums = jp.nodes(definitions, '$..[?(@.enum)]');
	/** @type {string[]} */
	const violations = [];
	for (const e of enums) {
		const x = e.value[X_ENUM_TYPE_NAME];
		if (!x) {
			const errMsg = `NO ${X_ENUM_TYPE_NAME} at: \`${e.path.join(".")}\`, value: \`${JSON.stringify(e.value)}\``;
			if (errMsg.includes("CANNOT_HANDLE")) {
				console.log("Safely Ignoring:", errMsg);
				continue;
			}
			violations.push(errMsg);
		}
	}
	return violations;
}

/**
 * @param {string} endpointUrl 
 */
async function validate(endpointUrl) {
	console.log("validate swagger endpoint:", endpointUrl);
	const res = await axios.get(endpointUrl);
	assert.strictEqual(res.status, 200, "should be http 200");
	const json = res.data;
	assert(json, "json falsy");
	/** @type {{[k:string]:SwaggerDefinition}} */
	const definitions = json.definitions;
	const v1 = validateInheritance(definitions);
	const v2 = validateEnumDefinition(definitions);
	const v3 = validateEnumInAllPositions(definitions);

	const violations = v1.concat(v2).concat(v3);
	if (violations.length) {
		const shouldSkip = Boolean(process.env[ENV_SKIP_SWAGGER_CHECK_KEY]);
		console.error(chalk.bgRed(`Swagger Definition ERRORS detected in ${endpointUrl}`));
		for (const v of violations) {
			console.error(chalk.red(v));
		}
		console.log(chalk.bgCyan(`To continue generating swagger regardless of the warnings above. Please set environment variable ${ENV_SKIP_SWAGGER_CHECK_KEY} to true and continue`));
		if (!shouldSkip) {
			process.exit(violations.length);
		}
	}
}

if (!module.parent) {
	(async () => {
		const endpointUrl = "http://www.mobiledevelopmentfht360web.com/swagger/docs/v1";
		// const endpointUrl = "http://localhost:2000/swagger/docs/v1";
		await validate(endpointUrl);
	})()
}

module.exports = {
	validate
};
